"""
This module contains functionality for reading statistics data from
the PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> s = pl.Statistics()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

_NONEXISTANT_VALUETYPE=-1
_NONEXISTANT_FIELD=-1
_NONEXISTANT_GRAPH_MODE=-1

class Priority(int):	
	priority = property(lambda self: self & 0x7f)
	expandable = property(lambda self: bool(self & 0x80))
	
class Statistics(Resource):
	"""The Statistics resource is used for for reading statistics.

	    @group Value types: VALUETYPE_ROOT, VALUETYPE_AGGREGATION, VALUETYPE_TRAFFICSTATS, VALUETYPE_ZONESTATS, VALUETYPE_CHANNELSTATS, VALUETYPE_NETOBJECT, VALUETYPE_HOST, VALUETYPE_LOCALVHOST, VALUETYPE_REMOTEVHOST, VALUETYPE_SERVOBJECT, VALUETYPE_SERVICE, VALUETYPE_CLIENTASPATH, VALUETYPE_SERVERASPATH, VALUETYPE_VLAN, VALUETYPE_INVLAN, VALUETYPE_OUTVLAN, VALUETYPE_INDSCP, VALUETYPE_OUTDSCP, VALUETYPE_INCHAN, VALUETYPE_OUTCHAN, VALUETYPE_INMPLS, VALUETYPE_OUTMPLS, VALUETYPE_XFBFLAG, VALUETYPE_CONNTAG, VALUETYPE_IPPROTOCOL, VALUETYPE_LINK, VALUETYPE_PROPERTY, VALUETYPE_BASESERVICE, VALUETYPE_ORIGINAS, VALUETYPE_INTBGPCOMM, VALUETYPE_EXTBGPCOMM, VALUETYPE_TTL, VALUETYPE_CHANNELSTATS_ROOT, VALUETYPE_CHANNELSTATS_CHANNEL
	    @group Total Fields: TOTALFIELD_BYTES_IN, TOTALFIELD_BYTES_OUT, TOTALFIELD_CONNS, TOTALFIELD_UCONNS, TOTALFIELD_CONNS_IN, TOTALFIELD_CONNS_OUT, TOTALFIELD_UCONNS_IN, TOTALFIELD_UCONNS_OUT, TOTALFIELD_QUALITY_IN, TOTALFIELD_QUALITY_OUT, TOTALFIELD_BYTES_TOTAL, TOTALFIELD_CURCONNS_IN, TOTALFIELD_CURCONNS_OUT, TOTALFIELD_PKTDROPS_IN, TOTALFIELD_PKTDROPS_OUT, TOTALFIELD_AVG_LATENCY_IN, TOTALFIELD_AVG_LATENCY_OUT, TOTALFIELD_SUBITEMS, TOTALFIELD_QUALITY_IN_INT, TOTALFIELD_QUALITY_OUT_INT, TOTALFIELD_QUALITY_IN_EXT, TOTALFIELD_QUALITY_OUT_EXT, TOTALFIELD_QUALITY_PKTS_IN, TOTALFIELD_QUALITY_PKTS_OUT, TOTALFIELD_RTT_IN, TOTALFIELD_RTT_OUT, TOTALFIELD_PACKETS_IN, TOTALFIELD_PACKETS_OUT
	    @group Graph Fields: GRAPHFIELD_BPS_IN, GRAPHFIELD_BPS_OUT, GRAPHFIELD_CPS, GRAPHFIELD_UCPS, GRAPHFIELD_CPS_IN, GRAPHFIELD_CPS_OUT, GRAPHFIELD_UCPS_IN, GRAPHFIELD_UCPS_OUT, GRAPHFIELD_CURCONNS_IN, GRAPHFIELD_CURCONNS_OUT, GRAPHFIELD_QUALITY_IN, GRAPHFIELD_QUALITY_OUT, GRAPHFIELD_SUBITEMS, GRAPHFIELD_BPS_TOTAL, GRAPHFIELD_PKTDROPS_IN, GRAPHFIELD_PKTDROPS_OUT, GRAPHFIELD_AVG_LATENCY_IN, GRAPHFIELD_AVG_LATENCY_OUT, GRAPHFIELD_QUALITY_IN_INT, GRAPHFIELD_QUALITY_OUT_INT, GRAPHFIELD_QUALITY_IN_EXT, GRAPHFIELD_QUALITY_OUT_EXT, GRAPHFIELD_QUALITY_PKTS_IN, GRAPHFIELD_QUALITY_PKTS_OUT, GRAPHFIELD_LINK_UTILIZATION_IN, GRAPHFIELD_LINK_UTILIZATION_OUT, GRAPHFIELD_RTT_IN, GRAPHFIELD_RTT_OUT, GRAPHFIELD_PACKETS_IN, GRAPHFIELD_PACKETS_OUT
	    @group Graph Modes: GRAPH_MODE_SIZE, GRAPH_MODE_SPEED
	    @group Channelstats Total Fields: TOTALFIELD_CHANNEL_RX_PACKETS, TOTALFIELD_CHANNEL_TX_PACKETS, TOTALFIELD_CHANNEL_RX_SPEED, TOTALFIELD_CHANNEL_TX_SPEED, TOTALFIELD_CHANNEL_RX_ERRORS, TOTALFIELD_CHANNEL_TX_ERRORS, TOTALFIELD_CHANNEL_RX_DROPS, TOTALFIELD_CHANNEL_TX_DROPS
	    @group Channelstats Graph Fields: GRAPHFIELD_CHANNEL_RX_PACKETS, GRAPHFIELD_CHANNEL_TX_PACKETS, GRAPHFIELD_CHANNEL_RX_SPEED, GRAPHFIELD_CHANNEL_TX_SPEED, GRAPHFIELD_CHANNEL_RX_ERRORS, GRAPHFIELD_CHANNEL_TX_ERRORS, GRAPHFIELD_CHANNEL_RX_DROPS, GRAPHFIELD_CHANNEL_TX_DROPS
	    @group Value flags: FLAG_VALID, FLAG_EXPANDABLE, FLAG_REMOTEPLDB, FLAG_GRAPHS, FLAG_PEAKS, FLAG_ISPEAK, FLAG_SYMLINK, FLAG_REMOTELINK,FLAG_IS_SUBSCRIBER, FLAG_HAS_SUBSCRIBERS

	    @cvar VALUETYPE_DUMMY: Root values (/ or Statistics Objects)
	    @cvar VALUETYPE_ROOT: Alias for VALUETYPE_DUMMY
	    @cvar VALUETYPE_AGGREGATION: Aggregation value
	    @cvar VALUETYPE_TRAFFICSTATS: Traffic stats value
	    @cvar VALUETYPE_ZONESTATS: Zone Stats value
	    @cvar VALUETYPE_CHANNELSTATS: Channel Stats value

	    @cvar VALUETYPE_NETOBJECT: Netobject
	    @cvar VALUETYPE_HOST: Host
	    @cvar VALUETYPE_LOCALVHOST: Local Vhost
	    @cvar VALUETYPE_REMOTEVHOST: Remote Vhost
	    @cvar VALUETYPE_SERVOBJECT: ServiceObject
	    @cvar VALUETYPE_SERVICE: Service
	    @cvar VALUETYPE_CLIENTASPATH: Client ASpath
	    @cvar VALUETYPE_SERVERASPATH: Server ASpath
	    @cvar VALUETYPE_INTASPATH: Internal ASpath
	    @cvar VALUETYPE_EXTASPATH: External ASpath
	    @cvar VALUETYPE_VLAN: Vlan
	    @cvar VALUETYPE_INVLAN: Inbound VLAN
	    @cvar VALUETYPE_OUTVLAN: Outbound VLAN
	    @cvar VALUETYPE_INDSCP: Inbound DSCP
	    @cvar VALUETYPE_OUTDSCP: Outbound DSCP
	    @cvar VALUETYPE_INCHAN: Inbound Channel
	    @cvar VALUETYPE_OUTCHAN: Outbound Channel
	    @cvar VALUETYPE_INMPLS: Inbound MPLS
	    @cvar VALUETYPE_OUTMPLS: Outbound MPLS
	    @cvar VALUETYPE_XFBFLAG: XFB Flags
	    @cvar VALUETYPE_CONNTAG: Categories
	    @cvar VALUETYPE_IPPROTOCOL: IP Protocol
	    @cvar VALUETYPE_REMOTEHOST: Remote Host
	    @cvar VALUETYPE_LINK: Link
	    @cvar VALUETYPE_PROPERTY: Property
	    @cvar VALUETYPE_BASESERVICE: Base Service
	    @cvar VALUETYPE_ORIGINAS: Origin AS
	    @cvar VALUETYPE_INTBGPCOMM: Internal BGP Community
	    @cvar VALUETYPE_EXTBGPCOMM: External BGP Community
	    @cvar VALUETYPE_TTL: Outbound TTL

	    @cvar VALUETYPE_CHANNELSTATS_ROOT: Channel Stats root
	    @cvar VALUETYPE_CHANNELSTATS_CHANNEL: Channel Stats channel

	    @cvar VALUETYPE_NAT_ROOT: NAT Stats Root
	    @cvar VALUETYPE_NAT_HOST: NAT Stats Host
	    @cvar VALUETYPE_NAT_ROOT: NAT Stats Pool

	    @cvar TOTALFIELD_BYTES_IN: Incoming Bytes count 
	    @cvar TOTALFIELD_BYTES_OUT: Outgoing Bytes count
	    @cvar TOTALFIELD_CONNS: Connections 
	    @cvar TOTALFIELD_UCONNS: Unestablished Connections 
	    @cvar TOTALFIELD_CONNS_IN: Incoming Connections
	    @cvar TOTALFIELD_CONNS_OUT: Outgoing Connections
	    @cvar TOTALFIELD_UCONNS_IN: Incoming unestablished Connections
	    @cvar TOTALFIELD_UCONNS_OUT: Outgoing unestablished Connections
	    @cvar TOTALFIELD_QUALITY_IN: Incoming Quality of Experience
	    @cvar TOTALFIELD_QUALITY_OUT: Outgoing Quality of Experience
	    @cvar TOTALFIELD_BYTES_TOTAL: Total bytes
	    @cvar TOTALFIELD_CURCONNS_IN: Incoming concurrent connections (Peak)
	    @cvar TOTALFIELD_CURCONNS_OUT: Incoming concurrent connections (Peak)
	    @cvar TOTALFIELD_PKTDROPS_IN: Incoming dropped packets
	    @cvar TOTALFIELD_PKTDROPS_OUT: Outgoing dropped packets
	    @cvar TOTALFIELD_AVG_LATENCY_IN: Incoming Avg Latency
	    @cvar TOTALFIELD_AVG_LATENCY_OUT: Outgoing Avg Latency
	    @cvar TOTALFIELD_SUBITEMS: Sub-Item Count
	    @cvar TOTALFIELD_QUALITY_IN_INT: Incoming Quality of Experience (Internal)
	    @cvar TOTALFIELD_QUALITY_OUT_INT: Outgoing Quality of Experience (Internal)
	    @cvar TOTALFIELD_QUALITY_IN_EXT: Incoming Quality of Experience (External)
	    @cvar TOTALFIELD_QUALITY_OUT_EXT: Outgoing Quality of Experience (External)
	    @cvar TOTALFIELD_QUALITY_PKTS_IN: Incoming Quality Packets
	    @cvar TOTALFIELD_QUALITY_PKTS_OUT: Outgoing Quality Packets
	    @cvar TOTALFIELD_RTT_IN: Internal Avg Handshake RTT
	    @cvar TOTALFIELD_RTT_OUT: External Avg Handshake RTT
	    @cvar TOTALFIELD_PACKETS_IN: Incoming Packets
	    @cvar TOTALFIELD_PACKETS_OUT: Outgoing Packets

	    @cvar GRAPHFIELD_BPS_IN: Incoming bytes/bps
	    @cvar GRAPHFIELD_BPS_OUT: Outgoing bytes/bps
	    @cvar GRAPHFIELD_CPS: Connections (per sec)
	    @cvar GRAPHFIELD_UCPS: Unestablished Connections (per sec)
	    @cvar GRAPHFIELD_CPS_IN: Incoming Connections (per sec)
	    @cvar GRAPHFIELD_CPS_OUT: Outgoing Connections (per sec)
	    @cvar GRAPHFIELD_UCPS_IN: Unestablished Incoming Connections (per sec)
	    @cvar GRAPHFIELD_UCPS_OUT: Unestablished Outgoing Connections (per sec)
	    @cvar GRAPHFIELD_CURCONNS_IN: Incoming concurrent connections
	    @cvar GRAPHFIELD_CURCONNS_OUT: Outgoing concurrent connections
	    @cvar GRAPHFIELD_QUALITY_IN: Incoming Quality of Experience
	    @cvar GRAPHFIELD_QUALITY_OUT: Outgoing Quality of Experience
	    @cvar GRAPHFIELD_SUBITEMS: Sub-Item Count
	    @cvar GRAPHFIELD_BPS_TOTAL: Total BPS
	    @cvar GRAPHFIELD_PKTDROPS_IN: Incoming Dropped Packets
	    @cvar GRAPHFIELD_PKTDROPS_OUT: Outgoing Dropped Packets
	    @cvar GRAPHFIELD_AVG_LATENCY_IN: Incoming Avg Latency
	    @cvar GRAPHFIELD_AVG_LATENCY_OUT: Outgoing Avg Latency
	    @cvar GRAPHFIELD_QUALITY_IN_INT: Incoming Quality of Experience (Internal)
	    @cvar GRAPHFIELD_QUALITY_OUT_INT: Outgoing Quality of Experience (Internal)
	    @cvar GRAPHFIELD_QUALITY_IN_EXT: Incoming Quality of Experience (External)
	    @cvar GRAPHFIELD_QUALITY_OUT_EXT: Outgoing Quality of Experience (External)
	    @cvar GRAPHFIELD_QUALITY_PKTS_IN: Incoming Quality Packets
	    @cvar GRAPHFIELD_QUALITY_PKTS_OUT: Outgoing Quality Packets
	    @cvar GRAPHFIELD_LINK_UTILIZATION_IN: Incoming Link Utlization
	    @cvar GRAPHFIELD_LINK_UTILIZATION_OUT: Outgoing Link Utlization
	    @cvar GRAPHFIELD_RTT_IN: Internal Avg Handshake RTT
	    @cvar GRAPHFIELD_RTT_OUT: External Avg Handshake RTT
	    @cvar GRAPHFIELD_PACKETS_IN: Incoming Packets
	    @cvar GRAPHFIELD_PACKETS_OUT: Outgoing Packets

	    @cvar TOTALFIELD_CHANNEL_RX_PACKETS: RX packets
	    @cvar TOTALFIELD_CHANNEL_TX_PACKETS: TX packets
	    @cvar TOTALFIELD_CHANNEL_RX_SPEED: RX speed
	    @cvar TOTALFIELD_CHANNEL_TX_SPEED: TX speed
	    @cvar TOTALFIELD_CHANNEL_RX_ERRORS: RX errors
	    @cvar TOTALFIELD_CHANNEL_TX_ERRORS: TX errors
	    @cvar TOTALFIELD_CHANNEL_RX_DROPS: RX drops
	    @cvar TOTALFIELD_CHANNEL_TX_DROPS: TX drops
	    @cvar GRAPHFIELD_CHANNEL_RX_PACKETS: RX packets
	    @cvar GRAPHFIELD_CHANNEL_TX_PACKETS: TX packets
	    @cvar GRAPHFIELD_CHANNEL_RX_SPEED: RX speed
	    @cvar GRAPHFIELD_CHANNEL_TX_SPEED: TX speed
	    @cvar GRAPHFIELD_CHANNEL_RX_ERRORS: RX errors
	    @cvar GRAPHFIELD_CHANNEL_TX_ERRORS: TX errors
	    @cvar GRAPHFIELD_CHANNEL_RX_DROPS: RX drops
	    @cvar GRAPHFIELD_CHANNEL_TX_DROPS: TX drops

	    @cvar GRAPH_MODE_SIZE: Graph result as absolute count
	    @cvar GRAPH_MODE_SPEED: Graph result as rate.

	    @cvar FLAG_VALID: Value is valid.
	    @cvar FLAG_EXPANDABLE: Value has values stored underneath.
	    @cvar FLAG_REMOTEPLDB: Unused
	    @cvar FLAG_GRAPHS: Value has graph data.
	    @cvar FLAG_PEAKS: Value has peak analysis data.
	    @cvar FLAG_ISPEAK: Value is a peak.
	    @cvar FLAG_SYMLINK: Value is a symbolic link.
	    @cvar FLAG_REMOTELINK: Value is a remote link.
	    @cvar FLAG_IS_SUBSCRIBER: Value is a subscriber.
	    @cvar FLAG_HAS_SUBSCRIBERS: Value has subscriber values stored underneath.

	"""

	VALUETYPE_DUMMY = PLDB.STATISTICS_VALUETYPE_ROOT
	VALUETYPE_ROOT = VALUETYPE_DUMMY
	VALUETYPE_AGGREGATION = PLDB.STATISTICS_VALUETYPE_AGGREGATION
	VALUETYPE_TRAFFICSTATS =  PLDB.STATISTICS_VALUETYPE_TRAFFICSTATS
	VALUETYPE_ZONESTATS =  PLDB.STATISTICS_VALUETYPE_ZONESTATS
	VALUETYPE_CHANNELSTATS = PLDB.STATISTICS_VALUETYPE_CHANNELSTATS

	VALUETYPE_NETOBJECT = PLDB.STATISTICS_VALUETYPE_NETOBJECT
	VALUETYPE_HOST = PLDB.STATISTICS_VALUETYPE_HOST
	VALUETYPE_LOCALVHOST = PLDB.STATISTICS_VALUETYPE_LOCALVHOST
	VALUETYPE_REMOTEVHOST = PLDB.STATISTICS_VALUETYPE_REMOTEVHOST
	VALUETYPE_SERVOBJECT = PLDB.STATISTICS_VALUETYPE_SERVOBJECT
	VALUETYPE_SERVICE = PLDB.STATISTICS_VALUETYPE_SERVICE
	VALUETYPE_INTASPATH = PLDB.STATISTICS_VALUETYPE_INTASPATH
	VALUETYPE_EXTASPATH = PLDB.STATISTICS_VALUETYPE_EXTASPATH
	VALUETYPE_VLAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_CLIENTASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_SERVERASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_INVLAN = PLDB.STATISTICS_VALUETYPE_INVLAN
	VALUETYPE_OUTVLAN = PLDB.STATISTICS_VALUETYPE_OUTVLAN
	VALUETYPE_INDSCP = PLDB.STATISTICS_VALUETYPE_INDSCP
	VALUETYPE_OUTDSCP = PLDB.STATISTICS_VALUETYPE_OUTDSCP
	VALUETYPE_INCHAN = PLDB.STATISTICS_VALUETYPE_INCHAN
	VALUETYPE_OUTCHAN = PLDB.STATISTICS_VALUETYPE_OUTCHAN
	VALUETYPE_INMPLS = PLDB.STATISTICS_VALUETYPE_INMPLS
	VALUETYPE_OUTMPLS = PLDB.STATISTICS_VALUETYPE_OUTMPLS
	VALUETYPE_XFBFLAG = PLDB.STATISTICS_VALUETYPE_XFBFLAG
	VALUETYPE_CONNTAG = PLDB.STATISTICS_VALUETYPE_CONNTAG
	VALUETYPE_IPPROTOCOL = PLDB.STATISTICS_VALUETYPE_IPPROTOCOL
	VALUETYPE_REMOTEHOST = PLDB.STATISTICS_VALUETYPE_REMOTEHOST
	VALUETYPE_LINK = PLDB.STATISTICS_VALUETYPE_LINK
	VALUETYPE_PROPERTY = PLDB.STATISTICS_VALUETYPE_PROPERTY
	VALUETYPE_BASESERVICE = PLDB.STATISTICS_VALUETYPE_BASESERVICE
	VALUETYPE_ORIGINAS = PLDB.STATISTICS_VALUETYPE_ORIGINAS
	VALUETYPE_INTBGPCOMM = PLDB.STATISTICS_VALUETYPE_INTBGPCOMM
	VALUETYPE_EXTBGPCOMM = PLDB.STATISTICS_VALUETYPE_EXTBGPCOMM
	VALUETYPE_TTL = PLDB.STATISTICS_VALUETYPE_TTL

	VALUETYPE_CHANNELSTATS_ROOT = PLDB.STATISTICS_VALUETYPE_CHANNELSTATS_ROOT
	VALUETYPE_CHANNELSTATS_CHANNEL = PLDB.STATISTICS_VALUETYPE_CHANNELSTATS_CHANNEL

	VALUETYPE_NAT_ROOT = PLDB.STATISTICS_VALUETYPE_NAT_ROOT
	VALUETYPE_NAT_HOST = PLDB.STATISTICS_VALUETYPE_NAT_HOST
	VALUETYPE_NAT_POOL = PLDB.STATISTICS_VALUETYPE_NAT_POOL

	_VALUETYPE_NAMES = {
		VALUETYPE_NETOBJECT: "NetObject",
		VALUETYPE_HOST: "Local Host",
		VALUETYPE_LOCALVHOST: "Local Vhost",
		VALUETYPE_REMOTEVHOST: "Remote Vhost",
		VALUETYPE_SERVOBJECT: "ServiceObject",
		VALUETYPE_SERVICE: "Service",
		VALUETYPE_INTASPATH: "Internal ASpath",
		VALUETYPE_EXTASPATH: "External ASpath",
		VALUETYPE_INVLAN: "In Vlan ID",
		VALUETYPE_OUTVLAN: "Out Vlan ID",
		VALUETYPE_INDSCP: "In DSCP",
		VALUETYPE_OUTDSCP: "Out DSCP",
		VALUETYPE_INCHAN: "In Channel",
		VALUETYPE_OUTCHAN: "Out Channel",
		VALUETYPE_INMPLS: "In MPLS",
		VALUETYPE_OUTMPLS: "Out MPLS",
		VALUETYPE_XFBFLAG: "XFB Flags",
		VALUETYPE_CONNTAG: "Categories",
		VALUETYPE_IPPROTOCOL: "IP Protocol",
		VALUETYPE_REMOTEHOST: "Remote Host",
		VALUETYPE_LINK: "Link",
		VALUETYPE_PROPERTY: "Property",
		VALUETYPE_BASESERVICE: "Base Service",
		VALUETYPE_ORIGINAS: "Origin AS",
		VALUETYPE_INTBGPCOMM: "Internal BGP Community",
		VALUETYPE_EXTBGPCOMM: "External BGP Community",
		VALUETYPE_TTL: "Outgoing TTL",

		VALUETYPE_CHANNELSTATS_ROOT: "Channel Statistics",
		VALUETYPE_CHANNELSTATS_CHANNEL: "Channel",

		VALUETYPE_NAT_ROOT: "NAT Statistics",
		VALUETYPE_NAT_HOST: "NAT Host",
		VALUETYPE_NAT_POOL: "NAT Pool",
	}

	TOTALFIELD_BYTES_IN = PLDB.STATISTICS_TOTALFIELD_BYTES_IN
	TOTALFIELD_BYTES_OUT = PLDB.STATISTICS_TOTALFIELD_BYTES_OUT
	TOTALFIELD_CONNS = PLDB.STATISTICS_TOTALFIELD_CONNS
	TOTALFIELD_UCONNS = PLDB.STATISTICS_TOTALFIELD_UCONNS
	TOTALFIELD_CONNS_IN = PLDB.STATISTICS_TOTALFIELD_CONNS_IN
	TOTALFIELD_CONNS_OUT = PLDB.STATISTICS_TOTALFIELD_CONNS_OUT
	TOTALFIELD_UCONNS_IN = PLDB.STATISTICS_TOTALFIELD_UCONNS_IN
	TOTALFIELD_UCONNS_OUT = PLDB.STATISTICS_TOTALFIELD_UCONNS_OUT
	TOTALFIELD_QUALITY_IN = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_OUT = _NONEXISTANT_FIELD
	TOTALFIELD_BYTES_TOTAL = PLDB.STATISTICS_TOTALFIELD_BYTES_TOTAL
	TOTALFIELD_CURCONNS_IN = PLDB.STATISTICS_TOTALFIELD_CURCONNS_IN
	TOTALFIELD_CURCONNS_OUT = PLDB.STATISTICS_TOTALFIELD_CURCONNS_OUT
	TOTALFIELD_PKTDROPS_IN = PLDB.STATISTICS_TOTALFIELD_PKTDROPS_IN
	TOTALFIELD_PKTDROPS_OUT = PLDB.STATISTICS_TOTALFIELD_PKTDROPS_OUT
	TOTALFIELD_AVG_LATENCY_IN = PLDB.STATISTICS_TOTALFIELD_AVG_LATENCY_IN
	TOTALFIELD_AVG_LATENCY_OUT = PLDB.STATISTICS_TOTALFIELD_AVG_LATENCY_OUT
	TOTALFIELD_SUBITEMS = PLDB.STATISTICS_TOTALFIELD_SUBITEMS
	TOTALFIELD_QUALITY_IN_INT = PLDB.STATISTICS_TOTALFIELD_QUALITY_IN_INT
	TOTALFIELD_QUALITY_OUT_INT = PLDB.STATISTICS_TOTALFIELD_QUALITY_OUT_INT
	TOTALFIELD_QUALITY_IN_EXT = PLDB.STATISTICS_TOTALFIELD_QUALITY_IN_EXT
	TOTALFIELD_QUALITY_OUT_EXT = PLDB.STATISTICS_TOTALFIELD_QUALITY_OUT_EXT
	TOTALFIELD_QUALITY_PKTS_IN = PLDB.STATISTICS_TOTALFIELD_QUALITY_PKTS_IN
	TOTALFIELD_QUALITY_PKTS_OUT = PLDB.STATISTICS_TOTALFIELD_QUALITY_PKTS_OUT
	TOTALFIELD_RTT_IN = PLDB.STATISTICS_TOTALFIELD_RTT_IN
	TOTALFIELD_RTT_OUT = PLDB.STATISTICS_TOTALFIELD_RTT_OUT
	TOTALFIELD_PACKETS_IN = PLDB.STATISTICS_TOTALFIELD_PACKETS_IN
	TOTALFIELD_PACKETS_OUT = PLDB.STATISTICS_TOTALFIELD_PACKETS_OUT

	GRAPHFIELD_BPS_IN = PLDB.STATISTICS_GRAPHFIELD_BPS_IN
	GRAPHFIELD_BPS_OUT = PLDB.STATISTICS_GRAPHFIELD_BPS_OUT
	GRAPHFIELD_CPS = PLDB.STATISTICS_GRAPHFIELD_CPS
	GRAPHFIELD_UCPS = PLDB.STATISTICS_GRAPHFIELD_UCPS
	GRAPHFIELD_CPS_IN = PLDB.STATISTICS_GRAPHFIELD_CPS_IN
	GRAPHFIELD_CPS_OUT = PLDB.STATISTICS_GRAPHFIELD_CPS_OUT
	GRAPHFIELD_UCPS_IN = PLDB.STATISTICS_GRAPHFIELD_UCPS_IN
	GRAPHFIELD_UCPS_OUT = PLDB.STATISTICS_GRAPHFIELD_UCPS_OUT
	GRAPHFIELD_CURCONNS_IN = PLDB.STATISTICS_GRAPHFIELD_CURCONNS_IN
	GRAPHFIELD_CURCONNS_OUT = PLDB.STATISTICS_GRAPHFIELD_CURCONNS_OUT
	GRAPHFIELD_QUALITY_IN = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_OUT = _NONEXISTANT_FIELD
	GRAPHFIELD_SUBITEMS = PLDB.STATISTICS_GRAPHFIELD_SUBITEMS
	GRAPHFIELD_BPS_TOTAL = PLDB.STATISTICS_GRAPHFIELD_BPS_TOTAL
	GRAPHFIELD_PKTDROPS_IN = PLDB.STATISTICS_GRAPHFIELD_PKTDROPS_IN
	GRAPHFIELD_PKTDROPS_OUT = PLDB.STATISTICS_GRAPHFIELD_PKTDROPS_OUT
	GRAPHFIELD_AVG_LATENCY_IN = PLDB.STATISTICS_GRAPHFIELD_AVG_LATENCY_IN
	GRAPHFIELD_AVG_LATENCY_OUT = PLDB.STATISTICS_GRAPHFIELD_AVG_LATENCY_OUT
	GRAPHFIELD_QUALITY_IN_INT = PLDB.STATISTICS_GRAPHFIELD_QUALITY_IN_INT
	GRAPHFIELD_QUALITY_OUT_INT = PLDB.STATISTICS_GRAPHFIELD_QUALITY_OUT_INT
	GRAPHFIELD_QUALITY_IN_EXT = PLDB.STATISTICS_GRAPHFIELD_QUALITY_IN_EXT
	GRAPHFIELD_QUALITY_OUT_EXT = PLDB.STATISTICS_GRAPHFIELD_QUALITY_OUT_EXT
	GRAPHFIELD_QUALITY_PKTS_IN = PLDB.STATISTICS_GRAPHFIELD_QUALITY_PKTS_IN
	GRAPHFIELD_QUALITY_PKTS_OUT = PLDB.STATISTICS_GRAPHFIELD_QUALITY_PKTS_OUT
	GRAPHFIELD_LINK_UTILIZATION_IN = PLDB.STATISTICS_GRAPHFIELD_LINK_UTILIZATION_IN
	GRAPHFIELD_LINK_UTILIZATION_OUT = PLDB.STATISTICS_GRAPHFIELD_LINK_UTILIZATION_OUT
	GRAPHFIELD_RTT_IN = PLDB.STATISTICS_GRAPHFIELD_RTT_IN
	GRAPHFIELD_RTT_OUT = PLDB.STATISTICS_GRAPHFIELD_RTT_OUT
	GRAPHFIELD_PACKETS_IN = PLDB.STATISTICS_GRAPHFIELD_PACKETS_IN
	GRAPHFIELD_PACKETS_OUT = PLDB.STATISTICS_GRAPHFIELD_PACKETS_OUT

	TOTALFIELD_CHANNEL_RX_PACKETS = PLDB.CHANNELSTATS_TFIELDS_RX_PACKETS
	TOTALFIELD_CHANNEL_TX_PACKETS = PLDB.CHANNELSTATS_TFIELDS_TX_PACKETS
	TOTALFIELD_CHANNEL_RX_SPEED = PLDB.CHANNELSTATS_TFIELDS_RX_SPEED
	TOTALFIELD_CHANNEL_TX_SPEED = PLDB.CHANNELSTATS_TFIELDS_TX_SPEED
	TOTALFIELD_CHANNEL_RX_ERRORS = PLDB.CHANNELSTATS_TFIELDS_RX_ERRORS
	TOTALFIELD_CHANNEL_TX_ERRORS = PLDB.CHANNELSTATS_TFIELDS_TX_ERRORS
	TOTALFIELD_CHANNEL_RX_DROPS = PLDB.CHANNELSTATS_TFIELDS_RX_DROPS
	TOTALFIELD_CHANNEL_TX_DROPS = PLDB.CHANNELSTATS_TFIELDS_TX_DROPS

	GRAPHFIELD_CHANNEL_RX_PACKETS = PLDB.CHANNELSTATS_GFIELDS_RX_PACKETS
	GRAPHFIELD_CHANNEL_TX_PACKETS = PLDB.CHANNELSTATS_GFIELDS_TX_PACKETS
	GRAPHFIELD_CHANNEL_RX_SPEED = PLDB.CHANNELSTATS_GFIELDS_RX_SPEED
	GRAPHFIELD_CHANNEL_TX_SPEED = PLDB.CHANNELSTATS_GFIELDS_TX_SPEED
	GRAPHFIELD_CHANNEL_RX_ERRORS = PLDB.CHANNELSTATS_GFIELDS_RX_ERRORS
	GRAPHFIELD_CHANNEL_TX_ERRORS = PLDB.CHANNELSTATS_GFIELDS_TX_ERRORS
	GRAPHFIELD_CHANNEL_RX_DROPS = PLDB.CHANNELSTATS_GFIELDS_RX_DROPS
	GRAPHFIELD_CHANNEL_TX_DROPS = PLDB.CHANNELSTATS_GFIELDS_TX_DROPS



	GRAPH_MODE_SIZE = PLDB.STATISTICS_GRAPH_MODE_SIZE
	GRAPH_MODE_SPEED = PLDB.STATISTICS_GRAPH_MODE_SPEED

	FLAG_VALID = PLDB.STATISTICS_FLAG_VALID
	FLAG_EXPANDABLE = PLDB.STATISTICS_FLAG_EXPANDABLE
	FLAG_REMOTEPLDB = PLDB.STATISTICS_FLAG_REMOTEPLDB
	FLAG_GRAPHS = PLDB.STATISTICS_FLAG_GRAPHS
	FLAG_PEAKS = PLDB.STATISTICS_FLAG_PEAKS
	FLAG_ISPEAK = PLDB.STATISTICS_FLAG_ISPEAK
	FLAG_SYMLINK = PLDB.STATISTICS_FLAG_SYMLINK
	FLAG_REMOTELINK = PLDB.STATISTICS_FLAG_REMOTELINK
	FLAG_IS_SUBSCRIBERS = PLDB.STATISTICS_FLAG_IS_SUBSCRIBER
	FLAG_HAS_SUBSCRIBERS =PLDB.STATISTICS_FLAG_HAS_SUBSCRIBERS

	_numtype = float

	def _resource(self):
		return PLDB.RESOURCE_STATREADER

	def _convert_valuetype_keys_to_name(self, d):
		r = {}
		for k, v in d.iteritems():
			r[self._VALUETYPE_NAMES.get(k, "Unknown value type %x" % k)] = v
		return r

	def _numtype_get(self):
		return self._numtype
	def _numtype_set(self, newval):
		self._numtype = newval
	_numtype_doc = """
	Set the type to use for numeric values in the result of statistics queries.

	Defaults to 'float', but as of firmware version 13.1 the data
	is stored with higher precision which may be lost when
	converting to a float. To get the increased precision in
	python this property can be set to int (or decimal.Decimal,
	but that is discouraged as that can easily create a huge
	amount of Decimal objects which consume much more memory than
	simple ints or floats).

		>>> s.list("2012-03-01", "2012-03-01", "/Applications?Statistics Object/Procera Categories?ServiceObject/")
		[{'flags': 523,
		  'name': 'File Sharing',
		  'subitemcounts': {'ServiceObject': 1.0},
		  'type': 517,
		  'values': {'bytes in': 75138613803.0,
			     'bytes out': 45238584181.0,
			     'bytes total': 120377197984.0}},
		 {'flags': 523,
		  'name': 'Streaming Media',
		  'subitemcounts': {'ServiceObject': 3.0},
		  'type': 517,
		  'values': {'bytes in': 24489327889.0,
		             'bytes out': 600116961.0,
		             'bytes total': 25089444850.0}}]

		>>> s.numtype=int
		>>> s.list("2012-03-01", "2012-03-01", "/Applications?Statistics Object/Procera Categories?ServiceObject/")[2]
		{'flags': 523,
		 'name': 'File Sharing',
		 'subitemcounts': {'ServiceObject': 1},
		 'type': 517,
		 'values': {'bytes in': 75138613803L,
			    'bytes out': 45238584181L,
			    'bytes total': 120377197984L}},
		 {'flags': 523,
		  'name': 'Streaming Media',
		  'subitemcounts': {'ServiceObject': 3},
		  'type': 517,
		  'values': {'bytes in': 24489327889L,
		             'bytes out': 600116961,
		             'bytes total': 25089444850L}}]


	@raises PLDBUnsupportedInFirmware: When trying to change the value in versions earlier than 13.1.
"""
	numtype = property(_numtype_get, _numtype_set, doc=_numtype_doc)

	def _top_list_post_process(self, nodelist):

		fields = [('bytes in', self.TOTALFIELD_BYTES_IN),
			  ('bytes out', self.TOTALFIELD_BYTES_OUT),
			  ('connections', self.TOTALFIELD_CONNS),
			  ('unest. connections', self.TOTALFIELD_UCONNS),
			  ('inbound connections', self.TOTALFIELD_CONNS_IN),
			  ('outbound connections', self.TOTALFIELD_CONNS_OUT),
			  ('inbound unest. connections', self.TOTALFIELD_UCONNS_IN),
			  ('outbound unest. connections', self.TOTALFIELD_UCONNS_OUT),
			  ('bytes total', self.TOTALFIELD_BYTES_TOTAL),
			  ('concurrent connections in', self.TOTALFIELD_CURCONNS_IN),
			  ('concurrent connections out', self.TOTALFIELD_CURCONNS_OUT),
			  ('shaping drops in', self.TOTALFIELD_PKTDROPS_IN),
			  ('shaping drops out', self.TOTALFIELD_PKTDROPS_OUT),
			  ('avg latency in', self.TOTALFIELD_AVG_LATENCY_IN),
			  ('avg latency out', self.TOTALFIELD_AVG_LATENCY_OUT),
			  ('subitems', self.TOTALFIELD_SUBITEMS),
			  ('in Quality internal', self.TOTALFIELD_QUALITY_IN_INT),
			  ('out Quality internal', self.TOTALFIELD_QUALITY_OUT_INT),
			  ('in Quality external', self.TOTALFIELD_QUALITY_IN_EXT),
			  ('out Quality externall', self.TOTALFIELD_QUALITY_OUT_EXT),
			  ('in Quality packets', self.TOTALFIELD_QUALITY_PKTS_IN),
			  ('out Quality packets', self.TOTALFIELD_QUALITY_PKTS_OUT),

]
		for e in nodelist:
			e[intern("subitemcounts")] = self._convert_valuetype_keys_to_name (e["subitemcounts"])

		ret = {}
		for field in fields:
			flist = []
			for x in nodelist:
				if x['values'].has_key(field[0]) and (x['topmask'] & field[1]):
					flist.append (x)

			if len(flist) == 0:
				continue

			valnam = field[0]
			def val_cmp(a,b):
				if a['values'][valnam] < b['values'][valnam]:
					return 1
				if a['values'][valnam] > b['values'][valnam]:
					return -1
				return 0
			flist.sort(val_cmp)

			ret['Toplist, %s' %field[0]] = flist

		for value, name in self._VALUETYPE_NAMES.iteritems():
			flist = []
			for x in nodelist:
				if x['subitemcounts'].has_key(name) and (x['topmask_subitem'] & (1 << (value - self.VALUETYPE_NETOBJECT))):
					flist.append (x)

			if len(flist) == 0:
				continue

			def si_cmp(a,b):
				if a['subitemcounts'][name] < b['subitemcounts'][name]:
					return 1
				if a['subitemcounts'][name] > b['subitemcounts'][name]:
					return -1
				return 0
			flist.sort(si_cmp)

			ret['Toplist, %s Subitems' % name] = flist

		return ret
	def date_list(self):
		"""
		List the statistics database of available statistics data.

			>>> x.date_list()
			['2005-12-01', '2005-12-02', '2005-12-03', '2005-12-04', '2005-12-05', '2005-12-06', '2005-12-07', '2005-12-08', '2005-12-09']

		@rtype: C{list} of C{str}
		@returns: a sorted list with dates that have statistics data

		"""

		lst = plutil._do(self, self.session.stat_date_list(), plutil._SimpleListCB(self, self.session.stat_date_list_handle))

		# Some sanity check that this seems to be a date and not e.g lost+found
		dates = [d for d in lst if len(d) == 10 and d[4] + d[7] == '--' and (d[0:4]+d[5:7]+d[-2:]).isdigit()]
		dates.sort()
		
		return dates

	def fields_list(self):
		"""
		List available fields.

			>>> s.fields_list()
			{'graphs': [{'desc': 'Incoming bps', 'id': 1, 'type': 1},
				    {'desc': 'Outgoing bps', 'id': 2, 'type': 1},
				    {'desc': 'CPS', 'id': 4, 'type': 2},
				    {'desc': 'Unestablished CPS', 'id': 8, 'type': 2},
				    {'desc': 'Incoming CPS', 'id': 16, 'type': 2},
				    {'desc': 'Outgoing CPS', 'id': 32, 'type': 2},
				    {'desc': 'Unestablished Incoming CPS', 'id': 64, 'type': 2},
				    {'desc': 'Unestablished Outgoing CPS', 'id': 128, 'type': 2},
				    {'desc': 'Incoming concurrrent connections', 'id': 256, 'type': 2},
				    {'desc': 'Outgoing concurrent connections', 'id': 512, 'type': 2},
				    {'desc': 'Incoming Quality of Experience', 'id': 1024, 'type': 3},
				    {'desc': 'Outgoing Quality of Experience', 'id': 2048, 'type': 3}],
			 'totals': [{'desc': 'Incoming Bytes', 'id': 1, 'type': 1},
				    {'desc': 'Outgoing Bytes', 'id': 2, 'type': 1},
				    {'desc': 'Connections', 'id': 4, 'type': 2},
				    {'desc': 'Unestablished Connections', 'id': 8, 'type': 2},
				    {'desc': 'Incoming Connections', 'id': 16, 'type': 2},
				    {'desc': 'Outgoing Connections', 'id': 32, 'type': 2},
				    {'desc': 'Incoming Unestablished Connections', 'id': 64, 'type': 2},
				    {'desc': 'Outgoing Unestablished Connections', 'id': 128, 'type': 2},
				    {'desc': 'Incoming Quality of Experience', 'id': 256, 'type': 3},
				    {'desc': 'Outgoing Quality of Experience', 'id': 512, 'type': 3}]}

		@returns: Dict with 'graphs' and 'totals'
		sub-dicts. The sub-dicts contain the keys 'desc', 'id'
		and 'type'. Type corresponds to the pathtype used in
		list/graph calls.
		
		@rtype: C{dict} of C{dict}
		"""
		r = plutil._do(self, self.session.stat_list_fields(), plutil._AddCB(self, self.session.stat_list_fields_handle))
		return r

	def list(self, start, end, path, pathtype=0, replymask=0):
		"""
		List stats nodes between two dates

			>>> s.list("2005-12-01", "2005-12-02", "/")
			[{'name': 'HostperNO', 
				'values': {'bytes out': 4900070621184.0,
					'connections': 351976704.0, 
					'outbound Quality of Experience': 99.520118713378906, 
					'inbound Quality of Experience': 99.921859741210938, 
					'bytes in': 21536893304832.0}, 
				'flags': 3,'type': 0}, 
			{'name': 'Top Sites', 
					'values': {'bytes out': 4900070621184.0, 
						'connections':351976704.0, 
						'bytes in':21536893304832.0},
					'flags': 3, 'type': 0}, 
			{'name': 'StatsByClass', 
					'values':{'bytes out': 4888053415936.0,
						'bytes in': 21408400801792.0},
					'flags': 3, 'type': 0}]

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: replymask parameter requires v14.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")
			
			
		cid = self.session.stat_list(start, end, path, pathtype, replymask)
		r = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_list_handle, (dict, self.numtype), False))

		for e in r:
			e[intern("subitemcounts")] = self._convert_valuetype_keys_to_name (e["subitemcounts"])

		return r
	
	def toplist(self, start, end, path, max, pathtype=0, replymask=0, offset=0):
		"""
		List top N nodes between two dates

		This function exists to limit the amount of data sent between 
		the server and the client.It returns any node which is in the 
		top N list for each of the fields being stored. 
		
		Also if the requested Statistics node contains sub-nodes of more
		than one type then a Top N list would be sent for each type for
		each field.

			>>> res=s.toplist("2009-01-14", "2009-01-14", "/v11 stats", 5, s.VALUETYPE_ROOT)
			>>> for node in res["Toplist, bytes in"]:
			...	if node['type'] == s.VALUETYPE_SERVICE:
			...     	print node
			... 
			{'values': {'bytes out': 30924881920.0, 'bytes in': 6812166656.0}, 'type': 518, 'name': 'BitTorrent transfer'}
			{'values': {'bytes out': 10508003328.0, 'bytes in': 2729902848.0}, 'type': 518, 'name': 'BitTorrent encrypted transfer'}
			{'values': {'bytes out': 2935312384.0, 'bytes in': 2098169216.0}, 'type': 518, 'name': 'HTTP'}
			{'values': {'bytes out': 577248896.0, 'bytes in': 1004788544.0}, 'type': 518, 'name': 'SMTP'}
			{'values': {'bytes out': 7716958720.0, 'bytes in': 923890496.0}, 'type': 518, 'name': 'CIFS'}


		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type max: C{int}
		@param max: Maximum number of nodes to output. If this is zero, all nodes will be returned.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@type offset: C{int}
		@param offset: Offset top list.

		@returns: Dict with the top list for each field stored ('In Bytes', 'Out Bytes', 
			'connections', 'Unest. Connections', 'In Connections',
			'Out Connections', 'In Quality of Experience' 'Out Quality of Experience'.
			Each list of is a list of dictonaries for each node.
		
		@rtype: C{dict} of C{list} of C{dict}

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: replymask parameter requires v14.0 firmware or newer.
		@attention: offset parameter requires v15.1 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_top_list(start, end, path, max, pathtype, replymask, offset)
		nodelist = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_top_list_handle, (dict, self.numtype), False))

		return self._top_list_post_process(nodelist)


	def list_duration(self, start, end, path, pathtype=0, thresholds=None):
		"""
		List nodes with calculated duration they satisfy thresholds.

			>>> s.list_duration("2010-09-08", "2010-09-08", "/All Host?/",thresholds=[(s.GRAPHFIELD_BPS_IN, 1e6,-1)])
			[{'duration': 36900L, 'flags': 11, 'name': '10.93.192.16', 'type': 514},
			 {'duration': 52500L, 'flags': 11, 'name': '10.93.192.17', 'type': 514},
			 {'duration': 300L, 'flags': 11, 'name': '10.93.192.25', 'type': 514},
			 {'duration': 49200L, 'flags': 11, 'name': '10.93.192.39', 'type': 514},
			 {'duration': 600L, 'flags': 11, 'name': '10.91.91.7', 'type': 514},
			 {'duration': 3900L, 'flags': 11, 'name': '10.92.97.2', 'type': 514},
			 {'duration': 300L, 'flags': 11, 'name': '10.92.97.5', 'type': 514},
			 {'duration': 300L, 'flags': 11, 'name': '10.92.97.10', 'type': 514},
			 {'duration': 1800L, 'flags': 11, 'name': '10.92.97.18', 'type': 514},
			 {'duration': 7500L, 'flags': 11, 'name': '10.92.97.25', 'type': 514},
			 {'duration': 3600L, 'flags': 11, 'name': '10.92.97.26', 'type': 514},
			 {'duration': 300L, 'flags': 11, 'name': '10.92.97.28', 'type': 514},
			 {'duration': 2400L, 'flags': 11, 'name': '10.91.91.12', 'type': 514},
			 {'duration': 3600L, 'flags': 11, 'name': '10.91.91.40', 'type': 514},
			 {'duration': 1800L, 'flags': 11, 'name': '10.91.91.41', 'type': 514},
			 {'duration': 69000L, 'flags': 11, 'name': '10.92.102.104', 'type': 514}]


		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type thresholds: C{list} of C{tuple}
		@param thresholds: List of tuples specifying thresholds. Each tuple is in the form I{(field, lower_limit, upper_limit)}, e.g: (s.GRAPHFIELD_BPS_TOTAL, 1e6, -1)

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If this function is called without firmware support.

		@attention: This function requires firmware v12.4 or newer
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_list_duration(start, end, path, pathtype, thresholds)
		r = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_list_duration_handle, dict, False))

		return r

	def _is_channelstats(self, path, pathtype):
		if pathtype in [self.VALUETYPE_CHANNELSTATS_ROOT, self.VALUETYPE_CHANNELSTATS_CHANNEL]:
			return True

		try:
			idx = path.rfind ('?')
		except ValueError:
			return False

		if path[-1] == '/':
			cmp_str = path[idx+1:-1]
		else:
			cmp_str = path[idx+1:]

		if cmp_str in [self._VALUETYPE_NAMES.get(self.VALUETYPE_CHANNELSTATS_ROOT),
			       self._VALUETYPE_NAMES.get(self.VALUETYPE_CHANNELSTATS_CHANNEL),
			       str(self.VALUETYPE_CHANNELSTATS_ROOT),
			       str(self.VALUETYPE_CHANNELSTATS_CHANNEL)]:
			return True

		return False

	def graph(self, start, end, path, numvals=24, pathtype=0, mode=1, replymask=0, usage_rate=False):
		"""
		Retrieve graph data between two dates.

			>>> s.graph("2009-01-05", "2008-01-05", "/Host per services", 24, 0, 1)
			[{'bytes out': 528960416.0, 'connections': 7485.5, 'bytes in': 95168704.0},
			 {'bytes out': 503229952.0, 'connections': 7073.0, 'bytes in': 136215120.0},
			 {'bytes out': 561612864.0, 'connections': 7457.75, 'bytes in': 165719856.0},
			 {'bytes out': 574073536.0, 'connections': 7316.66650390625, 'bytes in': 125461376.0}, 
			 {'bytes out': 575875264.0, 'connections': 7071.0, 'bytes in': 142020688.0}, 
			 {'bytes out': 533374208.0, 'connections': 7379.75, 'bytes in': 128362920.0}, 
			 {'bytes out': 542574656.0, 'connections': 7120.75, 'bytes in': 145863184.0}, 
			 {'bytes out': 627854016.0, 'connections': 7328.91650390625, 'bytes in': 124391448.0},
			 {'bytes out': 356714592.0, 'connections': 8949.1669921875, 'bytes in': 111401304.0}, 
			 {'bytes out': 276146240.0, 'connections': 9937.1669921875, 'bytes in': 121161912.0}, 
			 {'bytes out': 271035360.0, 'connections': 10208.8330078125, 'bytes in': 105619560.0}, 
			 {'bytes out': 282889088.0, 'connections': 9836.9169921875, 'bytes in': 105585656.0}, 
			 {'bytes out': 257417728.0, 'connections': 9915.75, 'bytes in': 118984736.0},
			 {'bytes out': 291532384.0, 'connections': 10910.6669921875, 'bytes in': 198847856.0}, 
			 {'bytes out': 281107264.0, 'connections': 10963.0830078125, 'bytes in': 96303560.0}, 
			 {'bytes out': 261479232.0, 'connections': 10322.6669921875, 'bytes in': 106611800.0}, 
			 {'bytes out': 365030240.0, 'connections': 10602.9169921875, 'bytes in': 233681280.0}, 
			 {'bytes out': 514466784.0, 'connections': 10718.75, 'bytes in': 129939368.0},
			 {'bytes out': 394594048.0, 'connections': 10602.5, 'bytes in': 128155672.0},
			 {'bytes out': 1116160000.0, 'connections': 11515.25, 'bytes in': 149660416.0},
			 {'bytes out': 423695584.0, 'connections': 10865.25, 'bytes in': 121420984.0},
			 {'bytes out': 592000384.0, 'connections': 10532.0, 'bytes in': 228060944.0},
			 {'bytes out': 623388928.0, 'connections': 9260.9169921875, 'bytes in': 198206720.0}, 
			 {'bytes out': 323936608.0, 'connections': 8981.25, 'bytes in': 117023752.0}]	

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type numvals: C{int}
		@param numvals: The number of data points to
		                retrieve. The specified time interval
		                is divided into this number of
		                segments and the average for each
		                segment is reported. In firmware
		                versions 13.1 and newer this value can
		                be set to 0 to retrieve all data
		                points that have been stored.

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type mode: C{int}
		@param mode: throughput rate or actual byte/connection count.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@type usage_rate: C{boolean}
		@param usage_rate: If usage_rate is set then graph rates are calculated based on micro usage bitmask. (requires firmwares v14.1 or newer).

		@rtype: C{dict} {list} of C{dict}
		@returns: Dictionaries with data.

		@raises PLDBUnsupportedInFirmware: If pathtype/mode/replymask/usage_rate arguments are used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: mode parameter requires v12.0 firmware or newer.
		@attention: replymask parameter requires v14.0 firmware or newer.
		@attention: usage_rate parameter requires v14.1 firmware or newer.
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		if mode == _NONEXISTANT_GRAPH_MODE:
			raise PLDBUnsupportedInFirmware("Specified 'mode' is not supported in this firmware version")

		cid = self.session.stat_graph(start, end, path, pathtype, numvals, mode, replymask, usage_rate)

		r = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_graph_handle, (dict, self.numtype), False))

		for e in r:
			if e.has_key("subitem count"):
				e[intern("subitem count")] = self._convert_valuetype_keys_to_name(e["subitem count"])
			
		return r

	def path_is_expandable(self, start, end, path, pathtype=0):
		"""
		Verify that a path is expandable.

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS
		
		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.  

		@returns: True if path is expandable.
		@rtype: C{boolean}

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_is_expandable (start, end, path, pathtype)
		if plutil._do(self, cid, plutil._AddCB(self, self.session.stat_is_expandable_handle)):
			return True
		return False

	def transfer_get(self, start, end, path, pathtype=0, replymask=0):
		"""
		Retrieves total transfer for path between two dates.

			>>> r.transfer_get("2006-01-20", "2006-01-21", "/NetObjects")
			{'priority': 129, 'out': 1458559580273L, 'name': '/NetObjects', 'in': 1632574443658L}

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@returns: Dictionary containing transfer information.
		@rtype: C{dict}

		@raises RuntimeError: B{"Could not retrieve data for path"}:
			If statistic path could not be read.
		@raises PLDBUnsupportedInFirmware: If pathtype argument is used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: replymask parameter requires v14.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_transfer_get (start, end, path, pathtype, replymask)
		res = plutil._do(self, cid, plutil._AddCB(self, self.session.stat_transfer_get_handle))
		#res["priority"] = Priority(res["priority"])
		return res

	def _is_writing(self):
		if plutil._do(self, self.session.stat_writer_status(), plutil._AddCB(self, self.session.stat_writer_status_handle)):
			return True
		return False

	def subscriber_count(self, start, end, path, pathtype=0):
		"""
		Count number of subscribers for requested path between the two dates

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@rtype: C{int}
		@returns: Integer count of subscribers.

		@raises PLDBUnsupportedInFirmware: If this function is called without firmware support.

		@attention: This function requires firmware v12.4 or newer
		"""
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")


		cid = self.session.stat_subscriber_count(start, end, path, pathtype)
		return plutil._do(self, cid, plutil._AddCB(self, self.session.stat_subscriber_count_handle))

	def subscriber_count_list(self, start, end, path, pathtype=0):
		"""
		List subscriber count for each node for requested path between the two dates

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If this function is called without firmware support.

		@attention: This function requires firmware v12.4 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")


		cid = self.session.stat_subscriber_count_list(start, end, path, pathtype)

		return plutil._do(self, cid, plutil._ListCB(self, self.session.stat_subscriber_count_list_handle, dict, False))

	def subscriber_count_graph(self, start, end, path, pathtype=0, numvals=24):
		"""
		Retrieve subscriber count for each requested interval between two dates.

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type numvals: C{int}
		@param numvals: The number of data points to
		                retrieve. The specified time interval
		                is divided into this number of
		                segments and the average for each
		                segment is reported. In firmware
		                versions 13.1 and newer this value can
		                be set to 0 to retrieve all data
		                points that have been stored.

		@rtype: C{list} of C{int}
		@returns: List of integers.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This function requires firmware v12.4 or newer
		"""
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_subscriber_count_graph(start, end, path, pathtype, numvals)
		return plutil._do(self, cid, plutil._ListCB(self, self.session.stat_subscriber_count_graph_handle, PLDB.PLSubscribercountGraph, False))

	def data_invalidate(self, start, end, path, pathtype=0, deleteroot=True):
		"""
		Invalidate all Statistics data for requested path and all its children

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type deleteroot: C{boolean}
		@param deleteroot: Invalidate data for requested path itself.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This method requires firmware v13.1 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		plutil._do(self, self.session.stat_data_invalidate(start, end, path, pathtype, deleteroot), plutil._AddCB(self))

	def graph_invalidate(self, start, end, path, pathtype=0, deleteroot=True):
		"""
		Invalidate only graph data for requested path and all its children

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type deleteroot: C{boolean}
		@param deleteroot: Invalidate graph data for requested path itself.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This method requires firmware v13.1 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		plutil._do(self, self.session.stat_graph_invalidate(start, end, path, pathtype, deleteroot), plutil._AddCB(self))

	def subitem_count(self, start, end, path, pathtype=0):
		"""
		Count number of subitems for requested path between the two dates

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@rtype: C{dict}
		@returns: Dictionary of various subtypes with their count.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This method requires firmware v13.1 or newer
		"""
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")


		cid = self.session.stat_subitem_count(start, end, path, pathtype)

		r = plutil._do(self, cid, plutil._AddCB(self, self.session.stat_subitem_count_handle))

		if r:
			return self._convert_valuetype_keys_to_name(r)

	def subitem_count_list(self, start, end, path, pathtype=0):
		"""
		Subitem count for the list of stats nodes between two dates

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This method requires firmware v13.1 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_subitem_count_list(start, end, path, pathtype)
		r = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_subitem_count_list_handle, dict, False))

		for e in r:
			e[intern("subitemcounts")] = self._convert_valuetype_keys_to_name (e["subitemcounts"])

		return r

	def list_subscribers(self, start, end, path, pathtype=0, max=0, replymask=0):
		"""
		List top N subscribers between two dates (N=0, returns all subscribers)

		This function exists to limit the amount of data sent between
		the server and the client.It returns any node which is in the
		top N list for each of the fields being stored.


			>>> res=s.list_subscribers("2013-09-14", "2013-09-14", "/Applications?Statistics Object")
			>>> for node in res:
			...     print node
			...
			{'values': {'bytes out': 30924881920.0, 'bytes in': 6812166656.0}, 'name': 'Sub113'}
			{'values': {'bytes out': 10508003328.0, 'bytes in': 2729902848.0},  'name': 'Sub5'}
			{'values': {'bytes out': 2935312384.0, 'bytes in': 2098169216.0},  'name': 'Sub49'}
			{'values': {'bytes out': 577248896.0, 'bytes in': 1004788544.0},  'name': 'Sub362'}
			{'values': {'bytes out': 7716958720.0, 'bytes in': 923890496.0},  'name': 'Sub74'}


		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@type max: C{int}
		@param max: Maximum number of nodes to output. If this is zero, all nodes will be returned.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@returns: If max=0, returns the list of items each stored as a Dict.
			If max is non-zero, Dict with the top list for each field
			stored or requested.  Each list of is a list of dictonaries
			for each node.

		@rtype: Max != 0, C{dict} of C{list} of C{dict}
		@rtype: Max == 0, C{list} of C{dict}

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: replymask parameter requires v14.0 firmware or newer.
		@attention: This method requires firmware v15.0.2 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_list_subscribers(start, end, path, pathtype, max, replymask)
		nodelist = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_list_subscribers_handle, (dict, self.numtype), False))

		if (max) :
			return self._top_list_post_process(nodelist)

		return nodelist

	def list_search(self, start, end, path, searchstr, pathtype = 0, max = 0, recursiondepth=0, replymask=0):
		"""
		List top N values matching a "search string" between two dates.
		N = 0, will return all the matching values.

			>>> res=s.list_search ("2013-10-02", "2013-10-02", "/Applications", "Sub115",)
			>>> for node in res:
			...     print node
			...
			{'values': {'bytes out': 30924881920.0, 'bytes in': 6812166656.0}, 'type': 518, 'name': '/Web Browsing?ServiceObject/HTTP?ServiceObject/Sub115?NetObject'}
			{'values': {'bytes out': 10508003328.0, 'bytes in': 2729902848.0}, 'type': 518, 'name': '/Messaging and Collaboration?ServiceObject/E-mail and Calendaring?ServiceObject/SMTP-TLS?ServiceObject/Sub115?NetObject'}
			{'values': {'bytes out': 2935312384.0, 'bytes in': 2098169216.0}, 'type': 518, 'name': '/Web Browsing?ServiceObject/HTTP download?ServiceObject/AN-115?NetObject'}


		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD or YYYY-MM-DD HH:MM:SS

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/NetObjects}

		@type searchstr: C{str}
		@param searchstr: The search string to match on.

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@type max: C{int}
		@param max: Maximum number of nodes to output. If this is zero, all nodes will be returned.

		@type recursiondepth: C{int}
		@param recursiondepth: Max depth of Statistics tree to search.

		@type replymask: C{int}
		@param replymask: If replymask is non-zero only data for the fields in the mask will be retured. (requires firmwares v14.0 or newer).

		@returns: If max=0, returns the list of items each stored as a Dict.
			If max is non-zero, Dict with the top list for each field
			stored or requested.  Each list of is a list of dictonaries
			for each node.

		@rtype: Max != 0, C{dict} of C{list} of C{dict}
		@rtype: Max == 0, C{list} of C{dict}


		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: This method requires firmware v15.0 or newer.
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_list_search(start, end, path, pathtype, max, searchstr, recursiondepth, replymask)
		nodelist = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_list_search_handle, (dict, self.numtype), False))

		if (max) :
			return self._top_list_post_process(nodelist)

		return nodelist

	def names_list(self, start, end, path, pathtype=0, matchstr=''):
		"""
		List the names of stats nodes between two dates

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list e.g. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type.

		@type path: C{str}
		@param path: Match string

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

		@attention: This method requires firmware v14.0.3 or newer
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)

		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_names_list(start, end, path, pathtype, matchstr)
		r = plutil._do(self, cid, plutil._ListCB(self, self.session.stat_names_list_handle, dict, False))

		return r
